using System;
using System.Collections;
using Cairo;

public class CuttingTree {
	public PointD[] Pieces;
	public string Chromosome;
	// 0 for normal, 1 for flipped
	public string Orientation;
	public TreeNode Root;
	
	/*public static void Main(string[] argv) {
		PointD[] pieces = new PointD[5];
		pieces[0] = new PointD(10, 10);
		pieces[1] = new PointD(20, 20);
		pieces[2] = new PointD(30, 30);
		pieces[3] = new PointD(40, 40);
		pieces[4] = new PointD(50, 50);
	
		CuttingTree tree = new CuttingTree(new PointD(200f, 200f), pieces, "1 2 H 3 V 4 H");
		
		Console.WriteLine((CuttingNode)tree._root);
	}*/
	
	public CuttingTree(PointD[] pieces, string chromosome, string orientation) {
		this.Pieces = pieces;
		this.Chromosome = chromosome;
		this.Orientation = orientation;
		
		this.Root = _buildTheTree();
	}
	
	private TreeNode _buildTheTree() {
		string tmp = "";
		
		// list of past nodes
		ArrayList nodes = new ArrayList();
		// last node of the all the nodes
		int last = 0;
		
		// create a simpler chromosome
		string chromosome = String.Concat(this.Chromosome, " ");
		
		foreach (Char c in chromosome) {
			if (c == ' ') {
				if (tmp != "H" && tmp != "V") {
					int pieceNumber = Int32.Parse(tmp);
					bool flipped = this.Orientation[pieceNumber] == '1';
					
					last = nodes.Add(new PieceNode(tmp, this.Pieces[pieceNumber], flipped));
				} else {
					CuttingNode current = new CuttingNode((TreeNode)nodes.get_Item(last-1), (TreeNode)nodes.get_Item(last));
					// Default is Horizontal
					if (tmp == "V")
						current.Method = CuttingMethod.Vertical;
					
					last = nodes.Add(current);
					// removes the used nodes
					nodes.RemoveRange(last-2, 2);
					last -= 2;
				}
				tmp = "";
			} else {
				tmp = String.Concat(tmp, c);
			}
		}
		
		return (TreeNode)nodes[last];
	}
}

public enum CuttingMethod : int {
	Horizontal,
	Vertical
}

// Node of the cutting tree
public abstract class TreeNode {
	public string Title;
	
	//--
	
	public TreeNode(String title) {
		this.Title = title;
	}
	
	public TreeNode() {
		this.Title = "";
	}
	
	public override abstract string ToString();
}

public class CuttingNode : TreeNode {
	
	private CuttingMethod _method;
	public CuttingMethod Method {
		get { return this._method; }
		set {
			this._method = value;
			if (value == CuttingMethod.Horizontal)
				this.Title = "H";
			else
				this.Title = "V";
		}
	}
	public TreeNode Left;
	public TreeNode Right;
	
	//--
	
	public CuttingNode(CuttingMethod method, TreeNode left, TreeNode right) {
		this.Method = method;
		this.Left = left;
		this.Right = right;
	}
	
	public CuttingNode(TreeNode left, TreeNode right)
		: this(CuttingMethod.Horizontal, left, right)
	{ }
	
	public CuttingNode(TreeNode left)
		: this(CuttingMethod.Horizontal, left, null)
	{ }
	
	public CuttingNode() : this(null, null)
	{ }
	
	//--
	
	public override string ToString() {
		return String.Format("{1} {2} {0}", this.Title, this.Left, this.Right);
	}
}

class PieceNode : TreeNode{
	private PointD _piece;
	private bool _flipped;
	
	public PointD Piece {
		get {
			return this._piece;
		}
		set {
			if (this.Flipped)
				this._piece = new PointD(value.Y, value.X);
			else
				this._piece = value;
		}
	}
	
	public bool Flipped {
		get {
			return this._flipped;
		}
		set {
			if (this._flipped != value) {
				this._piece = new PointD(this._piece.Y, this._piece.X);
			}
			this._flipped = value;
		}
	}
	
	public PieceNode(String title, PointD piece, bool flipped) : base(title) {
		this.Piece = piece;
		this.Flipped = flipped;
	}
	
	// Assuming that the piece isn't flipped by default
	public PieceNode(String title, PointD piece) : this(title, piece, false) {
	}
	
	public PieceNode() : base("")
	{ }
	
	public override string ToString() {
		return this.Title;
	}
}
