using Gtk;
using System;
using System.Collections;
using Cairo;

// http://www.go-mono.com/docs/index.aspx?link=T%3aGtk.DotNet.Graphics%2f*

/*

System for Sheet cutting optimization

*/

public class Genetic {
	public PointD Master;
	public PointD[] Pieces;
	
	public Genetic(PointD master, PointD[] pieces) {
		this.Master = master;
		this.Pieces = pieces;
	}
	
	// Generate a random tree
	public CuttingTree Generate() {
		string orientation = Genetic._generateOrientation(Pieces.Length);
		TreeNode root = Genetic._generateTree(Pieces, orientation);
		
		return new CuttingTree(this.Pieces, root.ToString(), orientation);
	}
	
	// Generate the random Orientation string
	private static string _generateOrientation(int size) {
		string val = "";
		Random rand = new Random();
		for (int i=0; i<size; i++) {
			if (rand.NextDouble() < .5)
				val = String.Concat(val, "0");
			else
				val = String.Concat(val, "1");
		}
		return val;
	}
	
	private static TreeNode _generateTree(PointD[] pieces, string orientation) {
		Random rand = new Random();
		// copy the array into an arraylist
		ArrayList p = new ArrayList(pieces.Length);
		p.AddRange(pieces);
		
		CuttingMethod method;
		CuttingNode cut;
		TreeNode left;
		TreeNode right;
		object left_element;
		object right_element;
		int left_index;
		int right_index;
		
		while(p.Count > 1) {
			left_index = rand.Next(p.Count);
			left_element = p.get_Item(left_index);
			p.RemoveAt(left_index);
			
			right_index = rand.Next(p.Count);
			right_element = p.get_Item(right_index);
			p.RemoveAt(right_index);
			
			if (left_element is PointD) {
				int index = Array.IndexOf(pieces, left_element);
				left = new PieceNode(String.Format("{0}", index), (PointD)left_element, orientation.get_Chars(index) == '1');
			} else {
				left = (TreeNode)left_element;
			}
			
			if (right_element is PointD) {
				int index = Array.IndexOf(pieces, right_element);
				right = new PieceNode(String.Format("{0}", index), (PointD)right_element, orientation.get_Chars(index) == '1');
			} else {
				right = (TreeNode)right_element;
			}
			
			// the cutting method
			if (rand.NextDouble() < .5)
				method = CuttingMethod.Horizontal;
			else
				method = CuttingMethod.Vertical;
			
			// cutting node
			cut = new CuttingNode(method, left, right);
			// add the new cutting node
			p.Add(cut);
		}
		
		return ((TreeNode)p.get_Item(0));
	}
}

public class GtkCairo {
	static CairoGraphic a;
	static Genetic genetic;
	
	static void Main () 	{		
		Application.Init ();
		Gtk.Window w = new Gtk.Window ("Cutting Essay");

		PointD[] pieces = new PointD[9];
		pieces[0] = new PointD(10f, 25f);
		pieces[1] = new PointD(12f, 23f);
		pieces[2] = new PointD(14f, 21f);
		pieces[3] = new PointD(16f, 19f);
		pieces[4] = new PointD(18f, 17f);
		pieces[5] = new PointD(20f, 15f);
		pieces[6] = new PointD(22f, 13f);
		pieces[7] = new PointD(24f, 11f);
		pieces[8] = new PointD(26f, 9f);
		
		genetic = new Genetic(new PointD(200, 200), pieces);
		
		CuttingTree tree = genetic.Generate();
		
		//CuttingTree tree = new CuttingTree(pieces, "0 1 V 2 H 3 4 V 5 6 V 7 V 8 H H V", "000001111");

		a = new CairoGraphic (tree);	
		
		Box box = new VBox (false, 0);
		Button button = new Button("Redraw");
		button.Clicked += redraw_clicked;
		box.Add (a);
		box.PackEnd (button, false, false, 0);
		w.Add (box);
		w.Resize (500,500);
		w.ShowAll ();		
		
		Application.Run ();	
	}
	
	static void redraw_clicked(object o, EventArgs args)
	{
		a.Tree = genetic.Generate();
		a.QueueDraw();
	}
}

public class CairoGraphic : DrawingArea {
   public CuttingTree Tree;
   
   public CairoGraphic(CuttingTree tree): base() {
   	this.Tree = tree;
   }
   
	public void draw (Cairo.Context gr, int width, int height)
	{
		gr.Scale(5, 5);
		gr.SetFontSize(2);
		gr.LineWidth = .2;
		
		drawNode(gr, this.Tree.Root);
		gr.Save();
		gr.Color = new Color(1, 0, 0, .5);
		gr.FillPreserve();
		gr.Color = new Color(0, 0, 0, .8);
		gr.Stroke();
	}
	
	static void drawNode(Cairo.Context gr, TreeNode node) {
		if (node is PieceNode) {
			drawNode(gr, (PieceNode)node);
		} else {
			drawNode(gr, (CuttingNode)node);
		}
	}
	
	static void drawNode(Cairo.Context gr, CuttingNode node) {
		PointD origin = gr.CurrentPoint;
		drawNode(gr, node.Left);
		PointD inter = gr.CurrentPoint;
		
		if (node.Method == CuttingMethod.Vertical) {
			gr.MoveTo(origin.X, gr.CurrentPoint.Y);
		} else {
			gr.MoveTo(gr.CurrentPoint.X, origin.Y);
		}
		
		drawNode(gr, node.Right);
		gr.MoveTo(Math.Max(gr.CurrentPoint.X, inter.X), Math.Max(gr.CurrentPoint.Y, inter.Y));
	}
	
	static void drawNode(Cairo.Context gr, PieceNode node) {
		gr.Rectangle(gr.CurrentPoint, node.Piece.X, node.Piece.Y);
		PointD origin = gr.CurrentPoint;
		gr.MoveTo(gr.CurrentPoint.X + 1, gr.CurrentPoint.Y + 3);
		gr.ShowText(String.Format("{0} ({1}, {2})", node.Title, node.Piece.X, node.Piece.Y));
		gr.MoveTo(origin);
		gr.MoveTo(gr.CurrentPoint.X + node.Piece.X, gr.CurrentPoint.Y + node.Piece.Y);
	}
   

	protected override bool OnExposeEvent (Gdk.EventExpose args)
	{
		Gdk.Window win = args.Window;
		//Gdk.Rectangle area = args.Area;
		
		Cairo.Context g = Gdk.Context.CreateDrawable (win);
		
		int x, y, w, h, d;
		win.GetGeometry(out x, out y, out w, out h, out d);
		
		draw (g, w, h);
		return true;
	}

}
